/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: db_struct.h

*/

/********************************/
/* Production System Components */
/********************************/

struct cpu_list {			/* List of available CPUs */

    char *cpu_name;			/* CPU Name */
    int cpu_type;			/* CPU Type */
    int task;				/* CPU Task Type, such as DB_TASK_IO */
    int opsys;				/* DB_VMS or DB_UNIX */
    int status;				/* DB_WORKING or DB_BROKEN */
    int power;				/* Power of the CPU, 100 = 1 VAX */
    int np;				/* Number of processes across all 
					   production systems */
    int refcount;			/* Number of references to this record
					   from other structures */
    struct cpu_list *next;		/* Next CPU in list */

};


struct fs_list {    			/* File system list */

    int id;	    			/* Internal ID of the file system */

    struct cpu_list *cpu;   /* CPU where the disk is physically located */

    char *mount_point;

    int mark;	    /* Extra field which is used in some algorithms to mark
		       that this filesystem has been seen, processed, etc. */

    int refcount;   /* Number of references to this filesystem */

    int blocks;	    /* Size, in blocks, of the filesystem */

    struct fs_list *next;

};

struct mount_info {

    char vsn[16];		/* Current tape on the drive */
				/* or current tape being requested */

    int rstate;			/* Ring state - DB_INPUT or DB_OUTPUT */
				/* of tape on drive or being requested */

    int mstate;			/* Mount state - DB_NOT_MOUNTED,
						 DB_MOUNT_PENDING,
						 DB_MOUNTED
				*/
};

struct tape_drive_list {		/* List of available tape drives */

    char *dev_name;			/* (Logical) Device Name (no dir) */
    char *physical;			/* Physical device path name */
    int dev_type;			/* Device Type */
    int status;				/* DB_WORKING or DB_BROKEN */
    struct cpu_list *cpu;		/* Pointer to CPU where tape drive
					   is located */
    struct mount_info *mi;		/* Mount information */
    int used;				/* DB_ALLOC or DB_UNALLOC */
    int refcount;			/* 0 = not part of a production system
					   1 = part of a production system */
    struct tape_drive_list *next;	/* Next tape drive in list */

};


/*******************/
/* Job Information */
/*******************/


struct word_list {

    char *word;
    struct word_list *next;
    struct word_list *down;	/* List within a list */

};


struct ordered_tape_list		/* Tape reel information */
{
    char *vsn;		    /* Volume serial number */

    long time_req;	    /* Time the tape was requested by a user process */

    long time_mnt;	    /* Time the tape was mounted */

    long time_done;	    /* Time the user declared he was finished with
			       the tape. */

    int status;		    /* DB_UNPROCESSED, DB_MOUNT_IN_PROG,
			       DB_CURRENT, DB_PROCESSED, DB_MISSING,
			       DB_DAMAGED */

    struct ordered_tape_list *next;  /* Next tape in the list */
};

struct tape_list_head {

    int listnum;			/* List Number */

    int devtype;			/* Device Type */

    int io;				/* 0 = read, -1 = write */

    int num_tapes;			/* Total number of tapes in the list */
    struct ordered_tape_list *tlist;

    struct tape_list_head *next;

};

struct class_list
{
    int class;				/* Class number */

    int opsys;				/* VMS or UNIX */

    struct tape_list_head *tlh;		/* Ordered tape lists */

    int debug;	    /* Number of processes to be started by hand */

    int minproc;    /* Minimum number of processes, -1 = all */

    int numproc;    /* Total number of processes, -1 = as many as possible */

    int percpu;	    /* #processes per CPU, -1 = all */

    int ptimeout;    /* -1 = not specified */

    int ctype;	    /* CPU type - -1 = not specified */

    struct word_list *scpu;    /* Specific CPUs */

    int share;	    /* Bit i is set if allowed to share with that class */

    int taptyp[4];  /* Type of tape drive (0 ends list of types being used) */
    int tapnum[4];  /* Number of drives needed for each type */

    struct word_list *exe;	/* Executable images (in a pipeline) 
				   includes arguments as a sublist for each
				   executable */

    char *stin;			/* File to redirect standard input to */

    char *stout;		/* File to redirect stdout (and stderr) to */

    struct word_list *dfils;	/* List of data files */

    char *outdir;		/* Output directory */

    struct word_list *outfil;	/* List of output files */

    struct class_list *next;

};

struct submit_info {

    char *jcf;		/* Name of the JCF file */
    char *jdf;		/* Name of the JDF file */
    char *username;	/* User that submitted the job */
    char *node;		/* Node job was submitted from */
    char *address;	/* Internet address of submit node */
    char *time;		/* Time the job was submitted */
};

struct prod_job_list {

    int entry;		    /* Number identifying the job */

    int pipestage;	    /* DB_PIPE_NEW, DB_PIPE_QUEUE, DB_PIPE_ON_DECK,
			       DB_PIPE_CURRENT, DB_PIPE_RUNDOWN */

    int status;		    /* DB_SUCCESS, DB_FAILURE, DB_ABORT,
			       DB_IN_PROGRESS, DB_PENDING
			       (this is the status at the current stage) */

    int jcss_pid;	    /* Pid of the jcss process */

    int area;		    /* Area the job is running in */

    int abort;		    /* 0 = no abort, -1 = abort job at next pipestage */

    int qpri;		    /* Queue Priority (lowest number has highest
			       priority) */

    struct submit_info *sub;

    char *on_deck_time;	    /* Times each phase started */
    char *start_time;
    char *rundown_time;

    struct class_list *clist;

    char *setup;    /* Command procedure, shell script, or executable to
		       run to set up the environment for the job */

    char *cleanup;  /* Command procedure, shell script or executable to
		       run on submitting machine after the job */

    char *logdir;		/* Log file directory */

    struct word_list *notif;	/* List of users to notify regardless of the
				   success of the job */

    struct word_list *notif_s;	/* List of users to notify if the job was
				   successful */

    struct word_list *notif_f;	/* List of users to notify if the job was
				   a failure */

    struct prod_job_list *next;

};

/*********************************/
/* Production System Definitions */
/*********************************/

struct prod_cpu_list {			/* Production System CPU List */

    struct cpu_list *cpu;		/* Pointer to CPU */
    int max_proc;			/* Maximum # Processes to run on CPU */

    int lognum;				/* Logical CPU Number */

    struct fs_list *fs;			/* Pointer to filesystem */
    char *mount_point;			/* Mount point on this CPU */

    int num_proc_left;			/* Number of process slots left
					   available when assigning specific
					   CPUs */

    struct prod_cpu_list *next;		/* Next CPU in list */

};

struct prod_tape_list {			/* Production System Tape Drive List */

    struct tape_drive_list *tape;	/* Pointer to tape drive */

    int used;				/* 0 = not used, -1 = used, 
					   (for assigning specific CPUs) */

    struct prod_tape_list *next;	/* Next tape drive in list */

};

struct prod_sys {			/* Production System Definition */

    char *ps_name;			/* Name of production system */

    char *account;			/* Account the job will run under */

    char *passwd;			/* Password for the account */

    char *login_dir;			/* Login directory of the account */

    struct cpu_list *jm_cpu;		/* Pointer to CPU where JM will run */

    struct fs_list *jm_fs;		/* Filesystem & mount point to copy */

    char *jm_mount_point;

    struct prod_cpu_list *cpus;		/* List of CPUs in this system */

    struct prod_tape_list *tapes;	/* List of tape drives */

    int status;				/*  0 = working, running (full power)
					   -1 = in repair, stopped
					    1 = working, reduced power */

    int next_entry;			/* Entry number for the next job
					   submitted */

    struct word_list *users;		/* List of authorized users */

    int next_area;
    int job_area[3];			/* Job areas */

    struct prod_job_list *newjob;	/* Record to hold job being submitted */

    struct prod_job_list *back;		/* Last job in queue */

    struct prod_job_list *job_queue;	/* List of jobs waiting to run.  The
					   first job on the list will be the
					   last job to run (jobs are inserted
					   at the front and taken from the
					   back) */

    struct prod_job_list *on_deck;	/* The job that is currently preparing
					   to run (or already prepared and
					   waiting to run).		    */

    struct prod_job_list *current;	/* Current job running on the
					   production system.  If no job is
					   currently running, this will be
					   NULL */

    struct prod_job_list *rundown;	/* The job that is currently in the
					   rundown phase. */

    struct prod_sys *next;		/* Pointer to next prod. system */

};

union itemptrs {
    char *str;
    struct fs_list *f;
    struct cpu_list *c;
    struct tape_drive_list *t;
    struct word_list *w;
    struct ordered_tape_list *otl;
    struct tape_list_head *tlh;
    struct class_list *cl;
    struct prod_job_list *pjl;
    struct prod_cpu_list *pcl;
    struct prod_tape_list *ptl;
    struct prod_sys *p;
    int i;
};

struct database {
    struct cpu_list *cpu;
    struct fs_list *fs;
    struct tape_drive_list *tape;
    struct prod_sys *ps;
};
