/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/****************************************************************************/
/*									    */
/*  Message Queue Utility Routines					    */
/*									    */
/*									    */
/*  IMPORTANT:  All operations on message queues should be done using 	    */
/*		these routines.  This centralizes the error handling.	    */
/*		All error messages are printed using the routine printerr.  */
/*									    */
/****************************************************************************/
/*									    */
/*  Message priorities:							    */
/*									    */
/*	Message priorities can range from 1 (highest priority) to 10	    */
/*  (lowest priority).  Messages should normally be sent with priority 5.   */
/*  The msgtyp parameter of the msq_put function should be set to the 	    */
/*  desired priority for each message sent.				    */
/*									    */
/****************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <errno.h>
#include "../inc/perr.h"

/*********************/
/* Macro Definitions */
/*********************/

#define FACILITY "MSQLIB"

/********************/
/* Global Variables */
/********************/

static char errmsg[132];

/**********************/
/* External Variables */
/**********************/

extern char	*msg_cmd, **msg_args;  /* Defined in MSGFMT.C */

/* #define diagnostics 0 */

/******************************************************************/
/*
    FUNCTION NAME:

      msq_put

  DESCRIPTION:

      Places a message on the specified message queue.  The message consists
      of a command string along with a variable number of argument strings.

  FORMAL PARAMETERS:

      int msqid - Message queue identifier.

      int msgtyp - Integer specifying the type of message.

      char *msgcmd - Command string.

      char **msgargs - List of argument strings terminated by a NULL
			    string.

  FUNCTION VALUE:

      int - returns zero if the message was successfully placed on the queue.

								  */
/******************************************************************/

int msq_put(msqid,msgtyp,msgcmd,msgargs)
int msqid;
long msgtyp;
char *msgcmd;
char **msgargs;
{

	/* Message queue information data structure (which is filled by a call
   to msgctl whenever the send fails */

	struct msqid_ds msqid_ds, *buf;

	/* Message size & flags */

	int msgsz;
	int msgflg = 0;

	/* Structure to hold the message contents and type */

	struct msgbuf1 {
		long mtype;
		char mtext[1024];
	} sndbuf, *msgp;

	buf = &msqid_ds;
	msgp = &sndbuf;

	/* Build a message and store it in the message structure */

	strcpy(msgp->mtext,msg_build(msgcmd,msgargs));
	msgp->mtype = msgtyp;

	/* Determine message size */

	msgsz = strlen(msgp->mtext) + 1;

	/* Send the message */

sndmsg:
	if (msgsnd(msqid,(struct msgbuf *)msgp,msgsz,msgflg) < 0) {	/* Send failed */

		/* Check to see if the send completed prematurely due to a signal */

		if (errno == EINTR) goto sndmsg;

		/* Print the reason for the failure */

		printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
		sprintf(errmsg,"message: \"%s\"",msgp->mtext);
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,2);

		/* If message queue was removed, return */

		if (errno == EIDRM) return(-1);

		/* Get message queue information and print it */

		if (msgctl(msqid,IPC_STAT,buf) == 0) {

			sprintf(errmsg,"message queue id = %d",msqid);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,3);
			sprintf(errmsg,"current # bytes on queue = %d",buf->msg_cbytes);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,4);
			sprintf(errmsg,"maximum # bytes on queue = %d",buf->msg_qbytes);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,5);
			sprintf(errmsg,"# messages on queue = %d",buf->msg_qnum);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,6);

		}
		else {
			printerr(FACILITY,ERROR_MSG,NULL,NULL,7);
		}

		return(-1);

	}

	/* Successful completion */

#ifdef diagnostics
	printf("MSQ_PUT: q = %d; m = \"%s\"; p = %ld\n",msqid,msgp->mtext,
	    msgp->mtype);
#endif
	return(0);

}

/******************************************************************/
/*
    FUNCTION NAME:

      msq_get

  DESCRIPTION:

      Removes a message from the specified message queue.  The message
      is decoded into a command string with a list of arguments.

  FORMAL PARAMETERS:

      int msqid - Message queue identifier.

      int msgtyp - Integer specifying the type of message desired.
	If msgtyp is less than zero, then the lowest type of message
	that is less than or equal to its absolute value is received.

  IMPLICIT OUTPUTS:

      char *msg_cmd (External) - Command string.

      char **msg_args (External) - List of argument strings terminated 
	    by a NULL string.

  FUNCTION VALUE:

      int - Returns the type (priority) of the message removed from the
	    queue.  If there was an error, -1 is returned.

								  */
/******************************************************************/

int msq_get(msqid,msgtyp)
int msqid;
long msgtyp;
{

	/* Message queue information data structure (which is filled by a call
   to msgctl whenever the send fails */

	struct msqid_ds msqid_ds, *buf;

	/* Message size & flags */

	int msgsz = 1024;
	int msgflg = 0;

	/* Structure to hold the message contents and type */

	static struct msgbuf1 {
		long mtype;
		char mtext[1024];
	} rcvbuf, *msgp;

	buf = &msqid_ds;
	msgp = &rcvbuf;

	/* Remove a message from the message queue */

rcvmsg:
	if (msgrcv(msqid,(struct msgbuf *)msgp,msgsz,msgtyp,msgflg) < 0) {

		/* Retry the read if it was aborted because of a signal */

		if (errno == EINTR) goto rcvmsg;

		/* Print out the reason for the failure */

		printerr(FACILITY,ERROR_MSG,NULL,NULL,8);

		if (errno == EIDRM) return(-1);

		/* Get message queue information and print it */

		if (msgctl(msqid,IPC_STAT,buf) == 0) {

			sprintf(errmsg,"message queue id = %d",msqid);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,9);
			sprintf(errmsg,"current # bytes on queue = %d",buf->msg_cbytes);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,10);
			sprintf(errmsg,"maximum # bytes on queue = %d",buf->msg_qbytes);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,11);
			sprintf(errmsg,"# messages on queue = %d",buf->msg_qnum);
			printerr(FACILITY,INFO_MSG,errmsg,NULL,12);

		}
		else {
			printerr(FACILITY,ERROR_MSG,NULL,NULL,13);
		}

		return(-1);

	}

#ifdef diagnostics
	printf("MSQ_GET: q = %d; m = \"%s\"; p = %ld\n",msqid,msgp->mtext,
	    msgp->mtype);
#endif

	/* Message has been successfully received, now we can parse it */

	msg_parse(msgp->mtext);

	/* Successful completion */

	return(msgp->mtype);

}

/******************************************************************/
/*
    FUNCTION NAME:

      msq_create

  DESCRIPTION:

      Creates a message queue.

  FORMAL PARAMETERS:

      none

  FUNCTION VALUE:

      int - message queue identifier if successful, otherwise -1.

								  */
/******************************************************************/

int msq_create()
{

	int msqid,msgflg;

	msgflg = 0666 | IPC_CREAT | IPC_EXCL;

	if ((msqid = msgget(IPC_PRIVATE,msgflg)) == -1)
		printerr(FACILITY,ERROR_MSG,NULL,NULL,14);

#ifdef diagnostics
	printf("MSQ_CREATE: message queue = %d\n",msqid);
#endif
	return(msqid);

}

/******************************************************************/
/*
    FUNCTION NAME:

      msq_delete

  DESCRIPTION:

      Removes the specified message queue from the system.

  FORMAL PARAMETERS:

      int msqid - message queue identifier.

  FUNCTION VALUE:

      int - returns zero if successful, -1 if not successful.

								  */
/******************************************************************/

int msq_delete(msqid)
int msqid;
{

	if (msgctl(msqid,IPC_RMID,0) < 0) {
		if (errno == EINVAL) return(0);	/* Queue is already deleted */
		printerr(FACILITY,ERROR_MSG,NULL,NULL,15);
		return(-1);
	}
	else {
#ifdef diagnostics
		printf("MSQ_DELETE: message queue = %d\n",msqid);
#endif
		return(0);
	}

}
