/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*                                                                      */
/*                                                                      */
/*			    Set Functions                               */
/*                                                                      */
/*                                                                      */
/*                                                                      */
/************************************************************************/

/*********************/
/* Macro Definitions */
/*********************/

#define BITS_TO_LONG(bits,nlong) nlong = bits/(8*sizeof(long)); \
				 if (nlong*8*sizeof(long) < bits) nlong++;


/******************************************************************/
/*
    FUNCTION NAME:

      create_empty_set

  DESCRIPTION:

      Allocates memory for a set and initializes it to zero (the empty set).

  FORMAL PARAMETERS:

      int size - Maximum size of the set.  This is the number of bits that
		 will be allocated.

  IMPLICIT INPUTS:

      none

  IMPLICIT OUTPUTS:

      none

  FUNCTION VALUE:

      (unsigned long *) - Pointer to memory allocated for the set.

								  */
/******************************************************************/

unsigned long *create_empty_set(size)
int size;
{
	int i,n;
	unsigned long *s;

	BITS_TO_LONG(size,n)
	    s = (unsigned long *) hmalloc(n*sizeof(long));  /* Allocate the memory */
	for (i=0; i<n; i++) s[i] = 0;		   /* Clear the memory */
	return(s);

}

set_clear(s,size)
unsigned long *s;
int size;
{
	int i,n;

	BITS_TO_LONG(size,n)
	    for (i=0; i<n; i++) s[i] = 0;		   /* Clear the memory */
}

/******************************************************************/
/*
    FUNCTION NAME:

      add_to_set

  DESCRIPTION:

      Adds one member to the set.

  FORMAL PARAMETERS:

      int member - Bit position which will be set to 1.

      unsigned long *set - Pointer to the bit string representing the set.

  IMPLICIT INPUTS:

      none

  IMPLICIT OUTPUTS:

      none

  FUNCTION VALUE:

      none
								  */
/******************************************************************/

add_to_set(member,set)
int member;
unsigned long *set;
{

	static unsigned long pattern[] = {  
		0x1,0x2,0x4,0x8,
		0x10,0x20,0x40,0x80,
		0x100,0x200,0x400,0x800,
		0x1000,0x2000,0x4000,0x8000,
		0x10000,0x20000,0x40000,0x80000,
		0x100000,0x200000,0x400000,0x800000,
		0x1000000,0x2000000,0x4000000,0x8000000,
		0x10000000,0x20000000,0x40000000,0x80000000 	};

	unsigned long i,bitpos;
	unsigned long x,y;

	x = member-1;
	y = 8*sizeof(unsigned long);
	i = x/y;
	bitpos = x - i*y;

	set[i] |= pattern[bitpos];

}

/******************************************************************/
/*
    FUNCTION NAME:

      remove_from_set

  DESCRIPTION:

      Removes a single element from a set.

  FORMAL PARAMETERS:

      int member - Bit position which will be set to 0.

      unsigned long *set - Pointer to the bit string representing the set.

  IMPLICIT INPUTS:

      none

  IMPLICIT OUTPUTS:

      none

  FUNCTION VALUE:

      none
								  */
/******************************************************************/

remove_from_set(member,set)
int member;
unsigned long *set;
{

	static unsigned long pattern[] = { 
		0xfffffffe,0xfffffffd,0xfffffffb,0xfffffff7,
		0xffffffef,0xffffffdf,0xffffffbf,0xffffff7f,
		0xfffffeff,0xfffffdff,0xfffffbff,0xfffff7ff,
		0xffffefff,0xffffdfff,0xffffbfff,0xffff7fff,
		0xfffeffff,0xfffdffff,0xfffbffff,0xfff7ffff,
		0xffefffff,0xffdfffff,0xffbfffff,0xff7fffff,
		0xfeffffff,0xfdffffff,0xfbffffff,0xf7ffffff,
		0xefffffff,0xdfffffff,0xbfffffff,0x7fffffff 	};

	unsigned long i,bitpos;
	unsigned long x,y;

	x = member-1;
	y = 8*sizeof(unsigned long);
	i = x/y;
	bitpos = x - i*y;

	set[i] &= pattern[bitpos];
}

/******************************************************************/
/*
    FUNCTION NAME:

      set_member

  DESCRIPTION:

      Tests to see if an element is a member of a set.

  FORMAL PARAMETERS:

      int member - The element to test.

      unsigned long *set - The set to test.

  IMPLICIT INPUTS:

      none

  IMPLICIT OUTPUTS:

      none

  FUNCTION VALUE:

      (int) - Returns 0 if not in the set, -1 if in the set.

								  */
/******************************************************************/

int set_member(member,set)
int member;
unsigned long *set;
{

	static unsigned long pattern[] = { 
		0x1,0x2,0x4,0x8,
		0x10,0x20,0x40,0x80,
		0x100,0x200,0x400,0x800,
		0x1000,0x2000,0x4000,0x8000,
		0x10000,0x20000,0x40000,0x80000,
		0x100000,0x200000,0x400000,0x800000,
		0x1000000,0x2000000,0x4000000,0x8000000,
		0x10000000,0x20000000,0x40000000,0x80000000 	};


	unsigned long i,bitpos;
	unsigned long x,y;

	x = member-1;
	y = 8*sizeof(unsigned long);
	i = x/y;
	bitpos = x - i*y;

	if ((set[i] & pattern[bitpos]) == 0)
		return(0);
	else
		return(-1);

}

/******************************************************************/
/*
    FUNCTION NAME:

      set_intersect

  DESCRIPTION:

      Returns the intersection of two sets.

  FORMAL PARAMETERS:

      unsigned long *set1,*set2 - The sets to intersect.

      int size - Size of each set.

  FUNCTION VALUE:

      (unsigned long *) - The intersection of the two sets.  The maximum set 
			  size allowed with this function is 
			  128*sizeof(unsigned long).

								  */
/******************************************************************/

unsigned long *set_intersect(set1,set2,size)
unsigned long *set1,*set2;
int size;
{
	static unsigned long iset[128];

	int i,nlong;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    for (i=0; i<nlong; i++) iset[i] = set1[i] & set2[i];

	return(iset);

}

/******************************************************************/
/*
    FUNCTION NAME:

      set_union

  DESCRIPTION:

      Performs set union operation.

  FORMAL PARAMETERS:

      unsigned long *set1,*set2 - set1 = set1 U set2.

      int size - Size of each set.

  FUNCTION VALUE:

      none.
								  */
/******************************************************************/

set_union(set1,set2,size)
unsigned long *set1,*set2;
int size;
{
	int i,nlong;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    for (i=0; i<nlong; i++) {
		*set1 |= *set2;
		set1++;
		set2++;
	}

	return;

}

/******************************************************************/
/*
    FUNCTION NAME:

      set_analyze

  DESCRIPTION:

      If the set contains one member, then the value of the member is returned.
      Zero is returned if the set is empty.
      -1 is returned if the set contains more than one member.

  FORMAL PARAMETERS:

      unsigned long *set - Set to search.

      int size - Number of possible members in the set.

								  */
/******************************************************************/

int set_analyze(set,size)
unsigned long *set;
int size;
{

	int nlong,i,j,ib,n;

	unsigned long m;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    ib = 1;
	j = 0;
	for (i=0; i<nlong; i++) {
		if (set[i] != 0) {
			m = set[i];
			n = 0;
			while(m != 0) {	/* Extract all members out of m */
				if ((m & 1) != 0) {
					if (j != 0) return(-1);
					j = ib+n;
				}
				n++;
				m >>= 1;
			}
		}
		ib += 8*sizeof(long);
	}
	return(j);

}

/******************************************************************/
/*
    FUNCTION NAME:

      set_query

  DESCRIPTION:

      Returns an array of int integers containing the members of
      the set.  The array is terminated by a zero.

  FORMAL PARAMETERS:

      unsigned long *set - Set to search.

      int size - Number of possible members in the set.

  FUNCTION VALUE:

      (int *) - Array of members of the set.

								  */
/******************************************************************/

int *set_query(set,size)
unsigned long *set;
int size;
{
	static int s[256];

	int nlong,i,j,ib,n;

	unsigned long m;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    j = 0;  /* Pointer into s */
	ib = 1;
	for (i=0; i<nlong; i++) {
		if (set[i] != 0) {
			m = set[i];
			n = 0;
			while(m != 0) {	/* Extract all members out of m */
				if ((m & 1) != 0) s[j++] = ib+n;
				n++;
				m >>= 1;
			}
		}
		ib += 8*sizeof(long);
	}
	s[j] = 0;
	return(s);

}

/******************************************************************/
/*
    FUNCTION NAME:

      set_query2 - added this for signal handlers to call.

  DESCRIPTION:

      Returns an array of int integers containing the members of
      the set.  The array is terminated by a zero.

  FORMAL PARAMETERS:

      unsigned long *set - Set to search.

      int size - Number of possible members in the set.

  FUNCTION VALUE:

      (int *) - Array of members of the set.

								  */
/******************************************************************/

int *set_query2(set,size)
unsigned long *set;
int size;
{
	static int s[256];

	int nlong,i,j,ib,n;

	unsigned long m;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    j = 0;  /* Pointer into s */
	ib = 1;
	for (i=0; i<nlong; i++) {
		if (set[i] != 0) {
			m = set[i];
			n = 0;
			while(m != 0) {	/* Extract all members out of m */
				if ((m & 1) != 0) s[j++] = ib+n;
				n++;
				m >>= 1;
			}
		}
		ib += 8*sizeof(long);
	}
	s[j] = 0;
	return(s);

}

/******************************************************************/
/*
    FUNCTION NAME:

      first_in_set

  DESCRIPTION:

      Returns the first item in the set, -1 if the set is empty.

  FORMAL PARAMETERS:

      unsigned long *set - The set to look at.

      int size - Number of possible members in the set.

  FUNCTION VALUE:

      Returns the first item in the set, -1 if the set is empty.

								  */
/******************************************************************/

int first_in_set(set,size)
unsigned long *set;
int size;
{

	int nlong,i,j,ib,n;

	unsigned long m;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    ib = 1;
	for (i=0; i<nlong; i++) {
		if (set[i] != 0) {
			m = set[i];
			n = 0;
			while(1) {
				if ((m & 1) != 0)
					if (size < (ib+n))
						return(-1);
					else
						return(ib+n);
				n++;
				m >>= 1;
			}
		}
		ib += 8*sizeof(long);
	}
	return(-1);

}

/******************************************************************/
/*
    FUNCTION NAME:

      first_not_in_set

  DESCRIPTION:

      Returns the first item not in the set, -1 if the set is full.

  FORMAL PARAMETERS:

      unsigned long *set - The set to look at.

      int size - Number of possible members in the set.

  FUNCTION VALUE:

      Returns the first item in the set, -1 if the set is full.

								  */
/******************************************************************/

int first_not_in_set(set,size)
unsigned long *set;
int size;
{

	int nlong,i,j,ib,n;

	unsigned long m;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    ib = 1;
	for (i=0; i<nlong; i++) {
		if (set[i] != -1) {
			m = set[i];
			n = 0;
			while(1) {
				if ((m & 1) == 0)
					if (size < (ib+n))
						return(-1);
					else
						return(ib+n);
				n++;
				m >>= 1;
			}
		}
		ib += 8*sizeof(long);
	}
	return(-1);
}

/******************************************************************/
/*
    FUNCTION NAME:

      set_size

  DESCRIPTION:

      Returns the number of members in the set.

  FORMAL PARAMETERS:

      unsigned long *set - Set to search.

      int size - Number of possible members in the set.

  FUNCTION VALUE:

    (int) number of members in the set.
								  */
/******************************************************************/

int set_size(set,size)
unsigned long *set;
int size;
{
	int z = 0;

	int nlong,i,j,ib,n;

	unsigned long m;

	BITS_TO_LONG(size,nlong)   /* Number of longwords in the set */

	    ib = 1;
	for (i=0; i<nlong; i++) {
		if (set[i] != 0) {
			m = set[i];
			n = 0;
			while(m != 0) {	/* Extract all members out of m */
				if ((m & 1) != 0) z++;
				n++;
				m >>= 1;
			}
		}
		ib += 8*sizeof(long);
	}
	return(z);
}
