/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File:   tape_io.c

    Explanation: 

    Exported Routines:


*/

#define FACILITY "TAPELIB"

#include "../inc/perr.h"

#include <stdio.h>
#include "../inc/types.h"
#include <sys/ioctl.h>
#include <errno.h>

#ifdef AIX
#include <sys/tape.h>
#else
#include <sys/mtio.h>
#endif

/*  General server routine to perform the IOCTL function set up by the
    specific entry points given below.  */

static int acp_tape_server(op, fildes, num)
int op, fildes, num;
{
#ifdef AIX
	struct stop tapeop;
#else
	struct mtop tapeop;
#endif
	int st = 0;

	int retry_count = 0;

#ifdef AIX
	tapeop.st_op = op;
	tapeop.st_count = num;
#else
	tapeop.mt_op = op;
	tapeop.mt_count = num;
#endif

retry:
#if defined(STIOCTOP)
	st = ioctl(fildes,STIOCTOP,&tapeop);
#endif
#if defined(MTIOCTOP)
	st = ioctl(fildes,MTIOCTOP,&tapeop);
#endif
	if (st < 0) {
		if (errno == EINTR) goto retry;
		if ((errno == EAGAIN) && (retry_count < 5)) {
			sleep_1(2);
			retry_count++;
			goto retry;
		}
		printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
		return(-1);
	}
	return(0);
}

/*  Skip forward 'num' files on the device with the given descriptor  */
int acp_skipfwd_file(fildes,num)
int fildes, num;
{
#if defined(STFSF)
	return acp_tape_server((int)STFSF,fildes,num);
#endif
#if defined(MTFSF)
	return acp_tape_server((int)MTFSF,fildes,num);
#endif
}

/*  Skip backward 'num' files on the device with the given descriptor  */
int acp_skipbck_file(fildes,num)
int fildes, num;
{
#if defined(STBSF)
	return acp_tape_server((int)STBSF,fildes,num);
#endif
#if defined(MTBSF)
	return acp_tape_server((int)MTBSF,fildes,num);
#endif
}

/*  Skip to END OF INFORMATION on the device with the given descriptor  */
int acp_skip_eoi(fildes)
int fildes;
{
#if defined(MTEOM)
	return acp_tape_server((int)MTEOM,fildes,1);
#endif
}

/*  Skip forward 'num' records on the device with the given descriptor  */
int acp_skipfwd_record(fildes,num)
int fildes, num;
{
#if defined(STFSR)
	return acp_tape_server((int)STFSR,fildes,num);
#endif
#if defined(MTFSR)
	return acp_tape_server((int)MTFSR,fildes,num);
#endif
}

/*  Skip backward 'num' records on the device with the given descriptor  */
int acp_skipbck_record(fildes,num)
int fildes, num;
{
#if defined(STBSR)
	return acp_tape_server((int)STBSR,fildes,num);
#endif
#if defined(MTBSR)
	return acp_tape_server((int)MTBSR,fildes,num);
#endif
}

/*  Write 'num' EOF marks on the device with the given descriptor  */
int acp_write_eof(fildes,num)
int fildes, num;
{
#if defined(STWEOF)
	return acp_tape_server((int)STWEOF,fildes,num);
#endif
#if defined(MTWEOF)
	return acp_tape_server((int)MTWEOF,fildes,num);
#endif
}

/*  Rewind the device with the given descriptor  */
int acp_rewind(fildes)
int fildes;
{
#if defined(STREW)
	return acp_tape_server((int)STREW,fildes,1);
#endif
#if defined(MTREW)
	return acp_tape_server((int)MTREW,fildes,1);
#endif
}

/*  Eject tape from device */
int acp_eject_tape(fildes)
int fildes;
{
#if defined(MTUNLOAD)
	return acp_tape_server((int)MTUNLOAD, fildes, 1);
#endif
}

/*  No-op to set the status of the device with the given descriptor  */
int acp_set_status(fildes)
int fildes;
{
#if defined(MTNOP)
	return acp_tape_server((int)MTNOP,fildes,1);
#endif
}

/*  Reset the device with the given descriptor  */
int acp_mt_reset(fildes)
int fildes;
{
#if defined(MTRST)
	return acp_tape_server((int)MTRST,fildes,1);
#endif
}

/************************************************************************/
/*

    Function:   rewind_unload

    Description: Rewind and unload a tape.

    Arguments:

    Function Value:  0 = success, -1 = failed.

*/
/************************************************************************/

int rewind_unload(fname)
char *fname;
{
	FILE *fp,*fopen();
	int fildes;
	int st;

	int retry_count = 0;

#ifdef SGI

	struct mtget devstat;

	/* Open the device file */

	fp = fopen(fname,"r");
	if (fp == NULL) return -1;
	fildes = fileno(fp);

	/* Get the device status block */

retry:
	st = ioctl(fildes,MTIOCGET,&devstat);
	if (st < 0) {
		if (errno == EINTR) goto retry;
		if ((errno == EAGAIN) && (retry_count < 5)) {
			sleep_1(2);
			retry_count++;
			goto retry;
		}
		fclose(fp);
		return -1;
	}

	/* Check to see if anything is in the tape drive */

	if (!devstat.mt_dposn) {	/* No tape is in the drive */
		fclose(fp);
		return -1;
	}
	if (acp_rewind(fildes) == -1) {
		fclose(fp);
		return -1;
	}
	if (acp_eject_tape(fildes) == -1) {
		fclose(fp);
		return -1;
	}
	fclose(fp);

#endif

	return 0;
}
