/*
**
** Demo program:
** 
** Uses the exponential map method to 
** integrate equations of motion of a 
** sextupolar type vector field.
** 
** As part of demo, we introduce the
** class field to make writing the
** vector field easier, as well as to
** illustrate style and elegance.
** 
** --- Leo Michelotti
** --- August 10, 1993
**
*/

#include <stdlib.h>
#include "mxyzptlk.rsc"

class field {
 private:
  Jet v[3];
 public:
  field( coord&, coord&, coord& );
  Jet operator[]( int );
};

field::field( coord& x, coord& y, coord& z ) {
 v[0] = - 2.0*x*y;
 v[1] =   x*x - y*y;
 v[2] =   0.0;
}

Jet field::operator[]( int i ) {
 return v[i];
}

main( int argc, char** argv ) {

 if( argc != 4 ) {
  printf( "\n" );
  printf( "Usage: %s  d  m  t                        \n", argv[0] );
  printf( "where  d  = highest order derivative      \n" );
  printf( "       m  = mass                          \n" );
  printf( "       t  = upper limit of integration.   \n" );
  printf( "\n" );
  exit(0);
 }

 int    deg = atoi( argv[1] );
 double msq = atof( argv[2] );
 double t   = atof( argv[3] );

 msq = msq*msq;
 Jet::setup( 6, deg, 6 );

 coord  x( 0.0 ),  y( 0.0 ),  z( 0.0 );
 coord px( 0.0 ), py( 0.0 ), pz( 0.0 );
 Jet E = sqrt( px*px + py*py + pz*pz + msq );
 field B( x, y, z );

 LieOperator v;
 v.setVariable(  px		       / E, 0 );
 v.setVariable(  py		       / E, 1 );
 v.setVariable(  pz		       / E, 2 );
 v.setVariable(  ( py*B[2] - pz*B[1] ) / E, 3 );
 v.setVariable(  ( pz*B[0] - px*B[2] ) / E, 4 );
 v.setVariable(  ( px*B[1] - py*B[0] ) / E, 5 );

 printf( "\n-*-*-*-  x  coefficients  -*-*-*-*-*\n" );
 v.expMap( t, x ).printCoeffs();
 printf( "\n-*-*-*-  y  coefficients  -*-*-*-*-*\n" );
 v.expMap( t, y ).printCoeffs();
 printf( "\n-*-*-*-  z  coefficients  -*-*-*-*-*\n" );
 v.expMap( t, z ).printCoeffs();

 LieOperator u, w, q;
 w.setVariable( x,  0 );
 w.setVariable( y,  1 );
 w.setVariable( z,  2 );
 w.setVariable( px, 3 );
 w.setVariable( py, 4 );
 w.setVariable( pz, 5 );
 w.reset();

 u = v.expMap( t, w );
 q = v.expMap( -t, w );

 printf( "\n-*-*-*-  u  coefficients  -*-*-*-*-*\n" );
 u.printCoeffs();

 printf( "\n-*-*-*-  u  coefficients  -*-*-*-*-*\n" );
 q.printCoeffs();
 
 printf( "\n-*-*-*-  u(q)  coefficients  -*-*-*-*-*\n" );
 u(q).printCoeffs();

 printf( "\n-*-*-*-  q(u)  coefficients  -*-*-*-*-*\n" );
 q(u).printCoeffs();
}
