#ifndef MATRIX_HXX
#define MATRIX_HXX

#include <complex.h>
#include "ML.hxx"

class MatrixC;
class ostream;

const complex complex_0 = complex(0.0,0.0);
const complex complex_1 = complex(1.0,0.0);

class MatrixD {
private:
  MLD* ml;

  // Friends
  friend class MatrixC;

  void copy_column(MatrixD& x,int from_col, int to_col);
  void switch_columns(int,int);
  void switch_rows(int,int);
  MatrixD scale();
  MatrixD lu_decompose(int*, int&);
  void lu_back_subst(int*,MatrixD&);
  void error(char * msgl, char * msg2 = " ")const ;

public:
  char  stacked;

  // Constructors and destructors_____________________________________
  MatrixD();
  MatrixD(int);
  MatrixD(int rows, int columns);
  MatrixD(int rows, int columns, double initval);
  MatrixD(int rows, int columns, double* initval);
  MatrixD(char* flag, int dimension); // create an identity matrix
				     // or a symplectic matrix
  MatrixD(const MatrixD& X);
  ~MatrixD();

  // Public member functions__________________________________________
  inline int rows() const { return ml->r;}
  inline int cols() const { return ml->c;}
  MatrixD transpose(); // transpose a square matrix
  double determinant();
  MatrixD inverse();
  MatrixC eigenValues();
  MatrixC eigenVectors();
  double trace(); // return the trace of a square matrix
  void SVD( MatrixD&, MatrixD&, MatrixD& ) const;
  

  // Operators________________________________________________________

  MatrixD& operator=(const MatrixD&);
  double& operator()(int row, int column);
  double& operator()(int row);
  void operator+=( const double&);
  void operator-=( const double&);
friend char operator==( const MatrixD&, const MatrixD& );
friend char operator==( const MatrixD&, const double& );
friend char operator==( const double&, const MatrixD& );
friend ostream& operator<<(ostream&, const MatrixD&);
//friend istream& operator>>(istream&, MatrixD&);
friend MatrixD operator+ (const MatrixD &,const MatrixD &);
friend MatrixD operator+ (const MatrixD &,const double&); 
friend MatrixD operator+ (const double& ,const MatrixD &); 
friend MatrixD operator-(const MatrixD &); 
friend MatrixD operator-(const MatrixD &,const MatrixD&); 
friend MatrixD operator-(const MatrixD &,const double&); 
friend MatrixD operator-(const double&,const MatrixD &); 
friend MatrixD operator*(const MatrixD &,const MatrixD&); 
friend MatrixC operator*(const MatrixD &,const MatrixC&); 
friend MatrixC operator*(const MatrixC &,const MatrixD&); 
friend MatrixD operator*(const MatrixD &,const double);
friend MatrixD operator*(const double,const MatrixD &);
friend MatrixD operator/(const MatrixD &,const double);
friend MatrixD operator/(const double,MatrixD &);
friend MatrixD operator/(MatrixD& ,MatrixD &);

};

void operator-=( MatrixD&, const MatrixD& );
void operator*=( MatrixD&, const MatrixD& );
//void operator/=( MatrixD&, const MatrixD& );
//void operator/=( MatrixD&, const double );
char operator!=( const MatrixD&, const MatrixD& );
char operator!=( const MatrixD&, const double& );
char operator!=( const double&, const MatrixD& );

class MatrixC {
private:
  MLC* ml;
  friend class MatrixD;


  void copy_column(MatrixC& x,int from_col, int to_col);
  void switch_columns(int,int);
  void switch_rows(int,int);
  MatrixC scale();
  MatrixC lu_decompose(int*, int&);
  void lu_back_subst(int*,MatrixC&);
  void error(char * msgl, char * msg2 = " ")const ;

public:
  char  stacked;

  // Constructors and destructors_____________________________________
  MatrixC();
  MatrixC(int);
  MatrixC(int rows, int columns );
  MatrixC(int rows, int columns, complex initval );
  MatrixC(int rows, int columns, complex* initval);
  MatrixC(const MatrixD&);
  MatrixC(char* flag, int dimension); // create an identity matrix
				     // or a symplectic matrix
  MatrixC(const MatrixC& X);
  ~MatrixC();

  // Public member functions__________________________________________
  inline int rows() const { return ml->r;}
  inline int cols() const { return ml->c;}
  MatrixC transpose(); // transpose a square matrix
  complex determinant();
  MatrixC inverse();
  MatrixC eigenValues();
  MatrixC eigenVectors();
  complex trace(); // return the trace of a square matrix
  

  // Operators________________________________________________________

  MatrixC& operator=(const MatrixC&);
  complex& operator()(int row, int column);
  complex& operator()(int row);
  void operator+=( const complex&);
  void operator-=( const complex&);
friend char operator==( const MatrixC&, const MatrixC& );
friend char operator==( const MatrixC&, const complex& );
friend char operator==( const complex&, const MatrixC& );
friend ostream& operator<<(ostream&, const MatrixC&);
//friend istream& operator>>(istream&, MatrixC&);
friend MatrixC operator+ (const MatrixC &,const MatrixC &);
friend MatrixC operator+ (const MatrixC &,const complex&); 
friend MatrixC operator+ (const complex& ,const MatrixC &); 
friend MatrixC operator-(const MatrixC &); 
friend MatrixC operator-(const MatrixC &,const MatrixC&); 
friend MatrixC operator-(const MatrixC &,const complex&); 
friend MatrixC operator-(const complex&,const MatrixC &); 
friend MatrixC operator*(const MatrixC &,const MatrixC&); 
friend MatrixC operator*(const MatrixC &,const MatrixD&); 
friend MatrixC operator*(const MatrixD &,const MatrixC&); 
friend MatrixC operator*(const MatrixC &,const complex);
friend MatrixC operator*(const complex,const MatrixC &);
friend MatrixC operator/(const MatrixC &,const complex);
friend MatrixC operator/(const complex,MatrixC &);
friend MatrixC operator/(const MatrixC &,const double);
friend MatrixC operator/(const double,MatrixC &);
friend MatrixC operator/(MatrixC& ,MatrixC &);
friend MatrixC operator/(MatrixD& ,MatrixC &);
friend MatrixC operator/(MatrixC& ,MatrixD &);

};

void operator-=( MatrixC&, const MatrixC& );
void operator*=( MatrixC&, const MatrixC& );
//void operator/=( MatrixC&, const MatrixC& );
//void operator/=( MatrixC&, const complex );
char operator!=( const MatrixC&, const MatrixC& );
char operator!=( const MatrixC&, const complex& );
char operator!=( const complex&, const MatrixC& );

class MatrixI {
private:
  MLI* ml;

  void error(char * msgl, char * msg2 = " ")const ;
public:
  char  stacked;

  // Constructors and destructors_____________________________________
  MatrixI();
  MatrixI(int);
  MatrixI(int rows, int columns );
  MatrixI(int rows, int columns, int initval );
  MatrixI(int rows, int columns, int* initval);
  MatrixI(char* flag, int dimension); // create an identity matrix
				     // or a symplectic matrix
  MatrixI(const MatrixI& X);
  ~MatrixI();

  // Public member functions__________________________________________
  inline int rows() const { return ml->r;}
  inline int cols() const { return ml->c;}
  MatrixI transpose(); // transpose a square matrix
  int trace();
  // Operators________________________________________________________

  MatrixI& operator=(const MatrixI&);
  int& operator()(int row, int column);
  int& operator()(int row);
  void operator+=( const int&);
  void operator-=( const int&);
friend char operator==( const MatrixI&, const MatrixI& );
friend char operator==( const MatrixI&, const int& );
friend char operator==( const int&, const MatrixI& );
friend ostream& operator<<(ostream&, const MatrixI&);
//friend istream& operator>>(istream&, MatrixI&);
friend MatrixI operator+ (const MatrixI &,const MatrixI &);
friend MatrixI operator+ (const MatrixI &,const int&); 
friend MatrixI operator+ (const int& ,const MatrixI &); 
friend MatrixI operator-(const MatrixI &); 
friend MatrixI operator-(const MatrixI &,const MatrixI&); 
friend MatrixI operator-(const MatrixI &,const int&); 
friend MatrixI operator-(const int&,const MatrixI &); 
friend MatrixI operator*(const MatrixI &,const MatrixI&); 
friend MatrixI operator*(const MatrixI &,const int);
friend MatrixI operator*(const int,const MatrixI &);
friend MatrixI operator/(const MatrixI &,const int);

};
 


#endif // MATRIX_HXX
