#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}


// ================================================================

#ifndef MX_SMALL
#define MX_SMALL    1.0e-12 // Used by JetC::addTerm to decide 
                            //   removal of a JetCterm.
#endif  // MX_SMALL

#ifndef MX_MAXITER
#define MX_MAXITER  100     // Maximum number of iterations allowed
                            //   in iterative routines
#endif  // MX_MAXITER




//***************************************************************
//***************************************************************
//***************************************************************
//
//    Implementation of class CLieOperator
//

//    Constructors and destructors    |||||||||||||||||||||||||||

CLieOperator::CLieOperator() {
 int i;
 
 if( JetC::oneEnv.SpaceDim == 0 ) {
   printf( "\n*** ERROR: CLieOperator::CLieOperator called " );
   printf( "after JetCSetup declared no vectors.\n" );
   exit(0);
   }
 
 weight = -1;
 accuWgt = JetC::oneEnv.MaxWeight;
 
 if( JetC::oneEnv.allSet == 137 ) {
   v = new JetC[ JetC::oneEnv.NumVar ];
   refPoint = new complex[ JetC::oneEnv.NumVar ];
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = JetC::oneEnv.refPoint[i];
   }
 else {
   v = 0;
   refPoint = 0;
   JetC::oneEnv.predefLieOp.insert( this );
   }
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator::CLieOperator( CLieOperator& x ) {
 int i;
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** CLieOperator::CLieOperator( CLieOperator& ) ERROR *** " );
   printf( "Prior call to JetCSetup required.\n" );
   exit(0);
   }
 
 v = new JetC[ JetC::oneEnv.NumVar ];

 for ( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   delete v[i].jl;
   v[i].jl = x.v[i].jl;    // Perform shallow copy.
   ( v[i]->rc )++;         // Increase reference counter.
 }

 weight = x.weight;
 accuWgt = x.accuWgt;
 
 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator::CLieOperator( LieOperator& x ) {
 int i;
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** CLieOperator::CLieOperator( CLieOperator& ) ERROR *** " );
   printf( "Prior call to JetCSetup required.\n" );
   exit(0);
   }
 
 v = new JetC[ JetC::oneEnv.NumVar ];
 for ( i = 0; i < JetC::oneEnv.NumVar; i++ ) v[i] = JetC(x.v[i]);
 weight = x.weight;
 accuWgt = x.accuWgt;
 
 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator::CLieOperator( JetC* x ) {    // A dangerous routine.
 int i;                                // It assumes the correct number
                                       // of components in the argument.

 static char firstCall = 1;

 if( firstCall ) {
  firstCall = 0;
  printf( "\n" );
  printf( "*** WARNING ***					\n" );
  printf( "*** WARNING *** CLieOperator::CLieOperator( JetC* x )	\n" );
  printf( "*** WARNING *** A dangerous routine.			\n" );
  printf( "*** WARNING *** It assumes the correct number	\n" );
  printf( "*** WARNING *** of components in the argument.	\n" );
  printf( "*** WARNING *** I highly recommend against its use!	\n" );
  printf( "*** WARNING ***					\n" );
 }

 if( JetC::oneEnv.SpaceDim == 0 ) {
  printf( "\n*** ERROR: CLieOperator::CLieOperator called " );
  printf( "after JetCSetup declared no vectors.\n" );
  exit(0);
 }
 
 if( JetC::oneEnv.allSet != 137 ) {
  printf( "\n*** CLieOperator::CLieOperator( CLieOperator& ) ERROR *** " );
  printf( "Prior call to JetCSetup required.\n" );
  exit(0);
 }
 
 weight = -1;
 accuWgt = JetC::oneEnv.MaxWeight;

 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x[0]->refPoint[i];

 v = new JetC[ JetC::oneEnv.NumVar ];
 for ( i = 0; i < JetC::oneEnv.NumVar; i++ ) setVariable( x[i], i );
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator::CLieOperator( char*  ) { 
 int i;
 int nv;
 
 if( JetC::oneEnv.SpaceDim == 0 ) {
   printf( "\n*** ERROR: CLieOperator::CLieOperator called " );
   printf( "after JetCSetup declared no vectors.\n" );
   exit(0);
   }
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** CLieOperator::CLieOperator( char* ) ERROR *** " );
   printf( "Prior call to JetCSetup required.\n" );
   exit(0);
   }
 
 nv = JetC::oneEnv.NumVar;
 v = new JetC[ nv ];

 refPoint = new complex[ nv ];
 for( i = 0; i < nv; i++ ) {
  refPoint[i] = JetC::oneEnv.refPoint[i];
  v[i]->setVariable( i );
 }

 weight = 1;
 accuWgt = JetC::oneEnv.MaxWeight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator::~CLieOperator() {
 if( v ) { delete [] v;
           delete [] refPoint; }
}

void CLieOperator::reset() {
 int i, j;
 weight = -1;
 accuWgt = JetC::oneEnv.MaxWeight;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
  if( weight < v[i]->weight ) weight = v[i]->weight;
  if( accuWgt > v[i]->accuWgt ) accuWgt = v[i]->accuWgt;
  refPoint[i] = v[0]->refPoint[i];
 }
 i = 1;
 while( i < JetC::oneEnv.NumVar ) {
  for( j = 0; j < JetC::oneEnv.NumVar; j++ ) 
   if( refPoint[j] != v[i]->refPoint[j] ) {
    printf( "\n" );
    printf( "*** ERROR ***                                      \n" );
    printf( "*** ERROR *** void CLieOperator::reset()            \n" );
    printf( "*** ERROR *** refPoint error for i = %d            \n", i );
    printf( "*** ERROR ***                                      \n" );
    exit(0);
   }
  i++;
 }
}


//     Member functions(public)  |||||||||||||||||||||||||||||||

void CLieOperator::fixReference( complex* r ) {
 int i, j;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = r[i];
 if( v ) {
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) v[j]->refPoint[i] = r[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::fixReference() {
 int i, j;

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  refPoint[i] = JetC::oneEnv.refPoint[i];

 if( v ) {
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = JetC::oneEnv.refPoint[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::fixReference( JetC& x ) {
 int i, j;

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  refPoint[i] = x->refPoint[i];

 if( v ) {
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = x->refPoint[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::fixReferenceAtEnd( const CLieOperator& x ) {
 int i, j;
 complex* u = new complex [ JetC::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = u[i];

 if( v ) {
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = u[i];
 }

 delete [] u;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::fixReferenceAtStart( const CLieOperator& x ) {
 int i, j;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];

 if( v ) {
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = x.refPoint[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::setVariable( const JetC& x, const int j ) {
 int i;
 
 if( j < 0 || JetC::oneEnv.NumVar <= j ) {
  printf( "\n***ERROR ***                                              \n" );
  printf(   "***ERROR *** CLieOperator::setVariable( JetC&, int )         \n" );
  printf(   "***ERROR *** It is certainly not true                     \n" );
  printf(   "***ERROR *** that 0 <= %d < %d                            \n",
            j, JetC::oneEnv.NumVar );
  printf(   "***ERROR ***                                              \n" );
  printf(   "***ERROR *** Sorry.  Good bye.                            \n" );
  printf(   "***ERROR ***                                              \n" );
  exit(0);
 }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) if( refPoint[i] != x->refPoint[i] ) {
  printf( "\n*** CLieOperator::setVariable ERROR: Incompatible references.\n" );
  exit(0);
 }
 
 v[j] = x;
 if ( weight  < x->weight  ) weight  = x->weight;
 if ( accuWgt > x->accuWgt ) accuWgt = x->accuWgt;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::setVariable( const complex x, const int j ) {
 int* ndx;
 int i;
 
 if( j < 0 || JetC::oneEnv.NumVar <= j ) {
  printf( "\n***ERROR ***                                              \n" );
  printf(   "***ERROR *** CLieOperator::setVariable( complex, int )         \n" );
  printf(   "***ERROR *** It is certainly not true                     \n" );
  printf(   "***ERROR *** that 0 <= %d < %d                            \n",
            j, JetC::oneEnv.NumVar );
  printf(   "***ERROR ***                                              \n" );
  printf(   "***ERROR *** Sorry.  Good bye.                            \n" );
  printf(   "***ERROR ***                                              \n" );
  exit(0);
 }
 
 v[j].clear();
 // if( ( j < 0 ) || ( j >= JetC::oneEnv.NumVar ) ) return;  ??? Old form.
 
 refPoint[j] = x;
 JetC::oneEnv.refPoint[j] = x;
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 ndx[j] = 1;
 v[j]->insert( new JLCterm( ndx, 1.0 ) );
 delete [] ndx;
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 v[j]->insert( new JLCterm( ndx, x ) );
 delete [] ndx;

 if ( weight < 1 ) weight = 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::getReference( complex* r ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) r[i] = refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char CLieOperator::isNilpotent() {
 for( int i = 0; i <  JetC::oneEnv.SpaceDim; i++ ) 
  if( !(v[i]->isNilpotent()) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::peekAt() {
 int i;
 printf( "\n******************\nCLieOperator::peekAt ********\n" );
 printf( "Weight: %d   Address: %d\n", weight, this );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   printf( "\n******************\n**** Component index = %d:\n", i );
   v[i]->peekAt();
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

dlist* CLieOperator::image() {
 dlist* ret = new dlist[ JetC::oneEnv.NumVar ];  // ??? Change this!!!
 for( int i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  ret[i] = v[i]->image();
 return ret;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::printCoeffs() {
 int i;
 cout << "\n************ Begin CLieOperator::printCoeffs ********\n"
      << "Weight: " << weight << endl;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   cout << "\n******************\n**** Component index = " << i << endl;
   v[i]->printCoeffs();
   }
 cout << "\n************ End   CLieOperator::printCoeffs ********" << endl;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::peekAt( char* fileName ) {
 int i;
 JetC::scratchFile = fopen( fileName, "w" );
 fprintf( JetC::scratchFile,
          "\n******************\nCLieOperator::peekAt ********\n" );
 fprintf( JetC::scratchFile, "Weight: %d   Address: %d\n", weight, this );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   fprintf( JetC::scratchFile,
            "\n******************\n**** Component index = %d:\n", i );
   v[i]->peekAt( 1 );
   }
 fclose( JetC::scratchFile );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::printCoeffs( char* fileName ) {
 int i;
 JetC::scratchFile = fopen( fileName, "w" );
 fprintf( JetC::scratchFile,
          "\n************ Begin CLieOperator::printCoeffs ********\n" );
 fprintf( JetC::scratchFile, "Weight: %d \n", weight );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   fprintf( JetC::scratchFile, "\n******************\n**** Component index = %d:\n", i );
   v[i]->printCoeffs( 1 );
   }
 fprintf( JetC::scratchFile, 
          "\n************ End   CLieOperator::printCoeffs ********\n" );
 fclose( JetC::scratchFile );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::standardPart( complex* x ) const {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )  x[i] = v[i]->standardPart();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::weightedDerivative( int* m, complex* x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) x[i] = v[i]->weightedDerivative( m );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void CLieOperator::derivative( int* m, complex* x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) x[i] = v[i]->derivative( m );
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::filter( MX_C_FUNCPTR f [] ) {
 CLieOperator z;
 int i;        

 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
  z.v[i] = v[i].filter( f[i] );

 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::filter( int lower, int upper ) {
 CLieOperator z;
 int i;        

 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
  z.v[i] = v[i].filter( lower, upper );

 z.reset();

 return z;
}


//     Operators   |||||||||||||||||||||||||||||||||||||||||||||

istream& operator>>(istream& is,  CLieOperator& x) {
 int i;
 int ival;
 complex cval;

  char buf[100];
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> x.weight;
  cout << " weight val = " << x.weight << endl;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> ival;
  is >> x.v[i];
 }
 return  is;
}

ostream& operator<<(ostream& os,  CLieOperator& x) {
 int i;
 os << "\n************ Begin CLieOperator  printCoeffs ********\n";
 os << "Weight: " << x.weight << endl;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   os << "\n******************\n**** Component index = " << i << endl;
   os << x.v[i];
   }
 return  os << "\n************ End   CLieOperator  printCoeffs ********\n";
}


char operator==( const CLieOperator& x, const CLieOperator& y ) {
 int i;
 if( x.weight  != y.weight  ) return 0;
 if( x.accuWgt != y.accuWgt ) return 0;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) if( x.refPoint[i] != y.refPoint[i] ) return 0;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) if( !( x.v[i] == y.v[i] ) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const CLieOperator& x, const CLieOperator& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char CLieOperator::operator==( const complex x ) {
 char answer = 1;
 for( int i = 0; i < JetC::oneEnv.SpaceDim; i++ ) 
  answer = answer && ( v[i] == x );
 return answer;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char CLieOperator::operator!=( const complex x ) {
 return !( this->operator==( x ) );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

MatrixC CLieOperator::Jacobian() const {
 int            i, j;
 int            nsd = JetC::oneEnv.SpaceDim;
 int            nv  = JetC::oneEnv.NumVar;
 int*           d = new int[ nsd ];
 MatrixC  M( nsd, nsd, 0.0 );

 for( i = 0; i < nv; i++  ) d[i] = 0;
 for( j = 0; j < nsd; j++ ) {
  d[j] = 1;
  for( i = 0; i < nsd; i++ ) M( i, j ) = v[i]->derivative( d );
  d[j] = 0;
 }

 delete [] d;
 return M;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::inverse() {          // ??? REMOVE: Obsolete

 static char firstTime = 1;
 if( firstTime ) {
  firstTime = 0;
  cout << "\n*** WARNING ***                                  \n"
          "*** WARNING *** You are using the obsolete         \n"
          "*** WARNING *** CLieOperator::inverse routine,     \n"
          "*** WARNING *** soon to vanish from the face of    \n"
          "*** WARNING *** the Earth.  Use the new and        \n"
          "*** WARNING *** improved CLieOperator::Inverse.    \n"
          "*** WARNING ***                                    \n"
       << endl;
 }

 CLieOperator z;
 CLieOperator  id( "ident" );                   // ??? Enlarge scope!!!
 CLieOperator  v;
 int          nsd  = JetC::oneEnv.SpaceDim;
 int          i = 0;
 MatrixC M( nsd, nsd, 0.0 );

 M = ( this->Jacobian() ).inverse();

 z = M*id;
 v = ( operator()(z) - id );
 while ( ( i++ < MX_MAXITER ) && ( v != 0.0 ) ) {
  z = z - M*v;
  v = ( operator()(z) - id );
 }

 if( i >= MX_MAXITER ) {
  printf( "*** WARNING ***                                        \n" );
  printf( "*** WARNING *** CLieOperator& CLieOperator::inverse()    \n" );
  printf( "*** WARNING *** Over %d iterations used;               \n", MX_MAXITER );
  printf( "*** WARNING *** result may be incorrect.               \n" );
  printf( "*** WARNING ***                                        \n" );
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::Inverse() const {

 if( JetC::oneEnv.SpaceDim <= 0 ) {
  cout << "*** ERROR ***                                     \n"
          "*** ERROR *** CLieOperator::Inverse               \n"
          "*** ERROR *** Prior call to JetC::Setup routine   \n"
          "*** ERROR *** required.                           \n"
          "*** ERROR ***                                     \n"
          "*** ERROR *** oneEnv.SpaceDim = " << JetC::oneEnv.SpaceDim << "\n"
          "*** ERROR ***                                     \n"
       << endl;
  exit(1);
 }


 int        i, j;
 char*      t1  = new char    [JetC::oneEnv.SpaceDim];
 char*      t2  = new char    [JetC::oneEnv.SpaceDim];
 char       t3;
 complex*   ref = new complex [JetC::oneEnv.NumVar];
 complex*  dref = new complex [JetC::oneEnv.NumVar];
 JLCterm**  p   = new JLCterm*[JetC::oneEnv.SpaceDim];


 // Store the default reference point

 for( j = 0; j < JetC::oneEnv.NumVar; j++ ) 
  dref[j] = JetC::oneEnv.refPoint[j];
 JetC::FixReference( JetC::oneEnv.AllZeroes ); 


 // If zero maps to zero, return result

 for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) {       // Actually, all
  t1[j] = 1;					      // componenets should
  for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )        // have the same
   t1[j] &= ( v[j]->refPoint[i] == complex_0 );       // reference point!
 }                                                    // ??? CHANGE THIS

 for( j = 0; j < JetC::oneEnv.SpaceDim; j++ )
  t2[j] = ( v[j].standardPart() == complex_0 );

 t3 = 1;
 for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) t3 &= ( t1[j] & t2[j] );
 if( t3 ) {
   JetC::FixReference( dref ); 
   delete [] t1;
   delete [] t2;
   delete [] ref;
   delete [] dref;
   delete [] p;

   return EpsInverse();
 }


 // Otherwise, construct an idempotent and compute its inverse.

 CLieOperator z;
 z = *this;   // Deep copy needed here.

 for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) {
  if( !t2[j] ) { 
   p[j] = z.v[j].get();
   z.v[j]->addTerm( new JLCterm( JetC::oneEnv.AllZeroes, complex_0 ) );
  }
 }

 if( !t1[0] ) 
  for( i = 0; i < JetC::oneEnv.SpaceDim; i++ ) {
   z.refPoint[i] = complex_0;
   for( j = 0; j < JetC::oneEnv.SpaceDim; j++ )   // A little paranoia
    z.v[j]->refPoint[i] = complex_0;              // never hurt.
  }

 z = z.EpsInverse();


 // Make final adjustments before returning.

 for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) {
  if( t2[j] ) ref[j] = complex_0;
  else        ref[j] = p[j]->value;
 }
 while( j < JetC::oneEnv.NumVar ) {
  ref[j] = refPoint[j];
  j++;
 }

 z.fixReference( ref );

 if( !t1[0] ) {   // ??? This should be sufficient.
  // A little check
  z.standardPart( ref );
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   if( ref[j] != complex_0 ) {
    cout << "*** WARNING ***                                              \n"
            "*** WARNING *** JetC::Inverse()                              \n"
            "*** WARNING *** ref[" << j << "] = " << ref[j] << "          \n"
            "*** WARNING ***                                              \n"
         << endl;
   }
  for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) 
   z.v[j]->addTerm( new JLCterm( JetC::oneEnv.AllZeroes, refPoint[j] ) );
 }


 // Clean up and leave

 JetC::FixReference( dref ); 
 delete [] t1;
 delete [] t2;
 delete [] ref;
 delete [] dref;
 for( j = 0; j < JetC::oneEnv.SpaceDim; j++ ) delete p[j];
 delete [] p;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::EpsInverse() const {
 CLieOperator z;
 CLieOperator  id( "ident" );
 CLieOperator  v;
 int          nsd  = JetC::oneEnv.SpaceDim;
 int          i = 0;
 MatrixC M( nsd, nsd, 0.0 );

 M = Jacobian().inverse();

 z = M*id;
 v = ( operator()(z) - id );
 while ( ( i++ < MX_MAXITER ) && ( ( v - v.filter(0,1)) != 0.0 ) ) {
               // This assumes linear is handled well enough
               // by the MatrixD routine.  
  z = z - M*v;
  v = ( operator()(z) - id );
 }

 if( i >= MX_MAXITER ) {
  cout << endl;
  cout << "*** WARNING ***                                            \n" 
          "*** WARNING *** CLieOperator& CLieOperator::EpsInverse()   \n" 
          "*** WARNING *** Over " << MX_MAXITER << " iterations used. \n"
          "*** WARNING *** result may be incorrect.                   \n" 
          "*** WARNING ***                                            \n" 
       << endl;
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::operator()( CLieOperator& x ) const {
 int i;
 CLieOperator z;
 for( i = 0; i < JetC::oneEnv.SpaceDim; i++) {
  z.v[i] = v[i]( x );
 }
 while( i < JetC::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator& CLieOperator::operator=( const CLieOperator& x ) {
 int i;
 int w;
 weight = -1;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   v[i] = x.v[i];
   if ( weight < ( w = v[i]->weight ) ) weight = w;
   refPoint[i] = x.refPoint[i];
   }
 accuWgt = x.accuWgt;
 return *this;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC CLieOperator::operator[]( int i ) {  // ??? Review the use of this???
 return v[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC CLieOperator::operator^( JetC& x ) { 
 JetC answer;
 int  s      = JetC::oneEnv.SpaceDim;
 int* ndx    = JetC::oneEnv.AllZeroes;
 int aw, i, jw;
 char adjustWeight = 1;

 aw = JetC::oneEnv.MaxWeight;
 ndx[0] = 1;
 answer = v[0]*x.D( ndx );
 ndx[0] = 0;

        // Differentiation will normally lower 
        //  the maximum accurate order by one.  This is 
        //  offset if the constant term in v[] vanishes.
        //
        if( v[0].standardPart() == complex_0 ) {
         jw = v[0]->accuWgt;
         if( jw > x->accuWgt ) jw = x->accuWgt;
         if( jw < aw ) aw = jw;
        }
        else adjustWeight = 0;

 if( s > 1 ) for( i = 1; i < s; i++ ) {
  ndx[i] = 1;
  answer += v[i]*x.D( ndx );
  ndx[i] = 0;

        if( v[i]->standardPart() == complex_0 ) {
         jw = v[i]->accuWgt;
         if( jw > x->accuWgt ) jw = x->accuWgt;
         if( jw < aw ) aw = jw;
        }
        else adjustWeight = 0;

 }

 // The accuracy of the answer is adjusted only if
 //  all components of v[] are differentials.
 // 
 if( adjustWeight ) answer->accuWgt = aw;

 answer.stacked = 1;
 return answer;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC CLieOperator::expMap( const double t, const JetC& x ) { 
 JetC answer;
 JetC u;
 double f;
 int count;
 
 answer = x;
 u      = x;
 f      = 1.0;
 count  = 0;

 u = ( t/f++ )*( operator^( u ) );
 while( ( count++ < MX_MAXITER ) && ( u != 0.0 ) ) {
  answer += u;                       // ??? How does this
  u = ( t/f++ )*( operator^( u ) );  // ??? affect efficiency?
 }

 if( count >= MX_MAXITER ) {
  printf( "\n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "*** WARNING *** JetC CLieOperator::expMap()               \n" );
  printf( "*** WARNING *** Number of iterations has exceeded %d    \n", MX_MAXITER );
  printf( "*** WARNING *** without achieving convergence.          \n" );
  printf( "*** WARNING *** Results may be incorrect.               \n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "\n" );
 }

 answer.stacked = 1;
 return answer;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC CLieOperator::expMap( const JetC& t, const JetC& x ) { 
 JetC answer;
 JetC u;
 double f;
 int count;
 
 answer = x;
 u      = x;
 f      = 1.0;
 count  = 0;

 u = ( t/f++ )*( operator^( u ) );
 while( ( count++ < MX_MAXITER ) && ( u != 0.0 ) ) {
  answer += u;                       // ??? How does this
  u = ( t/f++ )*( operator^( u ) );  // ??? affect efficiency?
 }

 if( count >= MX_MAXITER ) {
  printf( "\n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "*** WARNING *** JetC CLieOperator::expMap()               \n" );
  printf( "*** WARNING *** Number of iterations has exceeded %d    \n", MX_MAXITER );
  printf( "*** WARNING *** without achieving convergence.          \n" );
  printf( "*** WARNING *** Results may be incorrect.               \n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "\n" );
 }

 answer.stacked = 1;
 return answer;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::expMap( const complex t , const CLieOperator& x ) {
 CLieOperator z;

 for( int i = 0; i < JetC::oneEnv.SpaceDim; i++) 
  z.v[i] = expMap( t, x.v[i] );
 while( i < JetC::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator CLieOperator::expMap( const JetC& t, const CLieOperator& x ) {
 CLieOperator z;

 for( int i = 0; i < JetC::oneEnv.SpaceDim; i++) 
  z.v[i] = expMap( t, x.v[i] );
 while( i < JetC::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator+ ( const CLieOperator& x, const CLieOperator& y ) { 
 CLieOperator z;
 int i;

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( x.refPoint[i] != ( z.refPoint[i] = y.refPoint[i] ) ) {
     printf( "\n*** CLieOperator ADDITION ERROR: Inconsistency in reference points.\n" );
     exit(0);
     }
   }
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 z.weight = 0;
 z.accuWgt = JetC::oneEnv.MaxWeight;
 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  {
  z.v[i] = x.v[i] + y.v[i];
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
 }
 
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator- ( const CLieOperator& x, const CLieOperator& y ) {
 CLieOperator z;
 int i;

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( x.refPoint[i] != ( z.refPoint[i] = y.refPoint[i] ) ) {
     printf( "\n*** CLieOperator ADDITION ERROR: Inconsistency in reference points.\n" );
     exit(0);
     }
   }
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 z.weight = 0;
 z.accuWgt = JetC::oneEnv.MaxWeight;
 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  { 
  z.v[i] = x.v[i] - y.v[i];
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator- ( const CLieOperator& x ) {
 CLieOperator z;
 int i;

 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  z.v[i] = - x.v[i];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];
 z.weight  = x.weight;
 z.accuWgt = x.accuWgt;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator* ( const complex& c, const CLieOperator& x ) {
 CLieOperator z;
 int i;

 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  z.v[i] = c*x.v[i];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];
 z.weight  = x.weight;
 z.accuWgt = x.accuWgt;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator* ( const CLieOperator& x, const complex& c ) {
 CLieOperator z;
 int i;

 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  z.v[i] = c*x.v[i];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];
 z.weight  = x.weight;
 z.accuWgt = x.accuWgt;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator* ( const JetC& c, const CLieOperator& x ) {
 CLieOperator z;
 int i;

 z.weight = 0;
 z.accuWgt = JetC::oneEnv.MaxWeight;
 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  {
  z.v[i] = c*x.v[i];
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
  }

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator* ( const CLieOperator& x, const JetC& c ) {
 return c*x;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator*(  MatrixC& M, const CLieOperator& x ) {
 CLieOperator z;
 int i, j;

 z.weight = x.weight;
 z.accuWgt = x.accuWgt;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 int nsd = JetC::oneEnv.SpaceDim;

 if( ( M.rows() != nsd ) || 
     ( M.cols() != nsd )) {
  printf( "\n" );
  printf( "*** ERROR ***                                           \n" );
  printf( "*** ERROR *** CLieOperator& operator*(  MatrixC& M, CLieOperator& x ) \n" );
  printf( "*** ERROR *** Rows and/or columns of the matrix         \n" );
  printf( "*** ERROR *** are not correct.                          \n" );
  printf( "*** ERROR ***                                           \n" );
  exit(0);
 }

 complex temp;    // ??? I don't like this A LOT!!!
 for( i = 0; i < nsd; i++ ) {
   temp = M(i,0);
   z.v[i] = temp * x.v[0];
   j = 1;
   while( j < nsd ) {
     temp = M(i,j);
     z.v[i] += temp * x.v[j++];
   }
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

CLieOperator operator*( const CLieOperator& x, const CLieOperator& y ) { 
 CLieOperator z;
 CLieOperator u;
 JetC term;
 int i;
 int j;
 int w;
 JLCterm* p;
 
 // Check consistency of reference points.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   if( x.refPoint[i] != ( y.v[i] ).weightedDerivative( JetC::oneEnv.AllZeroes ) ) {
     printf( "\n*** CLieOperator CONCATENATION ERROR: Inconsistent reference points.\n" );
     exit(0);
     }
 
 // Subtract reference point prior to concatenation.
 u = y;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) { 
   p = ( u.v[i] ).get();
   if( p->weight != 0 ) {
     printf( "\n*** CLieOperator CONCATENATION ERROR: Argument has no nonstandard part!!\n" );
     exit(0);
   }
   delete p;
 }
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 JetC::oneEnv.JLmonomial[0] = 1.0;
 JetC::oneEnv.JLmonomial[0].fixReferenceAtStart( y );
 
 // For all higher weights ...
 for( w = 1; w <= x.weight; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, JetC::oneEnv.NumVar, Jet::oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( Jet::oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( Jet::oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = JetC::oneEnv.JLmonomial[ Jet::monoRank() ];
     ( Jet::oneEnv.exponent[i] )++;
     Jet::monoCode();
     JetC::oneEnv.JLmonomial[ Jet::monoRank() ] = term * u.v[i];   // ??? Is this OK???
 
     } // End while loop.
 
 // Monomials have been stored.
 // Now traverse the JetC variable and evaluate.
 
 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  {
   z.v[i] = concat( x.v[i] );
   for( j = 0; j < JetC::oneEnv.NumVar; j++ ) ( z.v[i] )->refPoint[j] = y.refPoint[j];
   }
 
 for( i = JetC::oneEnv.SpaceDim; i < JetC::oneEnv.NumVar; i++ ) {
   if( x.v[i] != y.v[i] )
     printf( "\n*** CLieOperator CONCATENATION ERROR!!! Horrible!!!\n" );
   z.v[i] = y.v[i];
   }
 
 // Set the correct reference point and weights.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) z.refPoint[i] = y.refPoint[i];
 
 z.weight = 0;
 z.accuWgt = JetC::oneEnv.MaxWeight;
 for( i = 0; i < JetC::oneEnv.SpaceDim; i++ )  {
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
 }
 
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void operator*=( CLieOperator& x, const CLieOperator& y ) {
 x = x*y;
}
